# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

from datahub.ingestion.graph.client import DataHubGraph


# Class used to fetch Secrets from DataHub.
class DataHubSecretsClient:
    _graph: DataHubGraph

    def __init__(self, graph):
        self._graph = graph

    def get_secret_values(self, secret_names: List[str]) -> Dict[str, Optional[str]]:
        if len(secret_names) == 0:
            return {}

        request_json = {
            "query": """query getSecretValues($input: GetSecretValuesInput!) {\n
                getSecretValues(input: $input) {\n
                    name\n
                    value\n
                }\n
            }""",
            "variables": {"input": {"secrets": secret_names}},
        }

        # Fetch secrets using GraphQL API f
        response = self._graph._session.post(
            f"{self._graph.config.server}/api/graphql", json=request_json
        )
        response.raise_for_status()

        # Verify response
        res_data = response.json()
        if "errors" in res_data:
            raise Exception("Failed to retrieve secrets from DataHub.")

        # Convert list of name, value secret pairs into a dict and return
        secret_value_list = res_data["data"]["getSecretValues"]
        secret_value_dict = dict()
        for secret_value in secret_value_list:
            secret_value_dict[secret_value["name"]] = secret_value["value"]
        return secret_value_dict



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Dict, List, Union

from acryl.executor.secret.secret_store import SecretStore


# Simple SecretStore implementation that fetches Secret values from the local environment.
class EnvironmentSecretStore(SecretStore):
    def __init__(self, config):
        pass

    def get_secret_values(self, secret_names: List[str]) -> Dict[str, Union[str, None]]:
        values = {}
        for secret_name in secret_names:
            values[secret_name] = os.getenv(secret_name)
        return values

    def get_secret_value(self, secret_name: str) -> Union[str, None]:
        return os.getenv(secret_name)

    def get_id(self) -> str:
        return "env"

    @classmethod
    def create(cls, config: Dict) -> "EnvironmentSecretStore":
        return cls(config)



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import time
from collections import OrderedDict
from datetime import datetime, timezone
from typing import List, Optional, Set

from datahub.configuration.common import OperationalError
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.graph.client import DatahubClientConfig, DataHubGraph
from datahub.metadata.schema_classes import (ExecutionRequestKeyClass,
                                             ExecutionRequestResultClass,
                                             StructuredExecutionReportClass)
from pydantic import validator

from acryl.executor.execution.default_executor import (DefaultExecutor,
                                                       DefaultExecutorConfig)
from acryl.executor.execution.task import TaskConfig
from acryl.executor.request.execution_request import (
    ExecutionProgressCallback, ExecutionRequest)
from acryl.executor.request.signal_request import SignalRequest
from acryl.executor.result.execution_result import ExecutionResult, Type
from acryl.executor.secret.secret_store import SecretStoreConfig

DATAHUB_EXECUTION_REQUEST_ENTITY_NAME = "dataHubExecutionRequest"
DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME = "dataHubExecutionRequestResult"
REPORTS_TO_EMIT_MAX_SIZE = 10


logger = logging.getLogger(__name__)


class ReportingExecutorConfig(DefaultExecutorConfig):
    id: str = "default"
    task_configs: List[TaskConfig]
    secret_stores: List[SecretStoreConfig]

    graph_client: Optional[DataHubGraph] = None
    graph_client_config: Optional[DatahubClientConfig] = None

    class Config:
        arbitrary_types_allowed = True

    @validator("graph_client")
    def check_graph_connection(cls, v):
        if v is not None:
            v.test_connection()
        return v


# Supports RUN_INGEST commands.
class ReportingExecutor(DefaultExecutor):
    _datahub_graph: DataHubGraph
    _execution_timeout_sec: Optional[float] = 86400  # Default timeout is 1 day

    def __init__(self, exec_config: ReportingExecutorConfig) -> None:
        super().__init__(exec_config)
        # Stores execution results in case completion mcp can't be emitted
        # Implements LRU cache via OrderedDict to avoid memory leak
        self.results_to_emit: OrderedDict[str, ExecutionResult] = OrderedDict()
        self.dropped_reports: Set[str] = set()  # Leaks memory slowly, like task_futures
        if exec_config.graph_client is not None:
            self._datahub_graph = exec_config.graph_client
        elif exec_config.graph_client_config is not None:
            self._datahub_graph = DataHubGraph(exec_config.graph_client_config)
        else:
            raise Exception(
                "Invalid configuration provided. Missing DataHub graph client configs"
            )

    # Run a list of tasks in sequence
    def execute(self, request: ExecutionRequest) -> ExecutionResult:
        # Capture execution start time
        request.start_time = datetime.now(timezone.utc)

        # Build & emit an ACK mcp
        kickoff_mcp = self._build_kickoff_mcp(request)
        self._datahub_graph.emit_mcp(kickoff_mcp, async_flag = False)

        # Execute the request
        request.progress_callback = self._make_progress_callback(request)
        exec_result = super().execute(request)

        if request.exec_id is not None:
            self.results_to_emit[request.exec_id] = exec_result
            if len(self.results_to_emit) > REPORTS_TO_EMIT_MAX_SIZE:
                exec_id, _ = self.results_to_emit.popitem(last=False)
                self.dropped_reports.add(exec_id)

        completion_mcp = self._build_completion_mcp(exec_result)
        # Additional retry logic as we really want to emit here, for total of 15 minutes
        for retry in range(6):
            if retry:
                time.sleep(60 * retry)
            if self._emit_completion_mcp(request.exec_id, completion_mcp):
                break

        return exec_result

    def signal(self, request: SignalRequest) -> None:
        super().signal(request)
        # First, try to emit stored completion MCP
        # If that doesn't exist and the task cannot be found (is not actively executing)
        # emit an empty cancellation event
        if request.signal == "KILL":
            exec_id = request.exec_id
            should_emit_cancellation = (
                exec_id not in self.task_futures or exec_id in self.dropped_reports
            )

            exec_result = self.results_to_emit.get(exec_id)
            if exec_result:
                completion_mcp = self._build_completion_mcp(exec_result)
                if self._emit_completion_mcp(exec_id, completion_mcp):
                    return
                else:
                    should_emit_cancellation = True

            if should_emit_cancellation:
                # No task found. Simply emit a cancelled MCE. The start time is unclear.
                # Build & emit the cancellation mcp
                cancellation_mcp = self._build_empty_cancel_mcp(exec_id=exec_id)
                self._datahub_graph.emit_mcp(cancellation_mcp, async_flag = False)

    def _make_progress_callback(
        self, exec_request: ExecutionRequest
    ) -> ExecutionProgressCallback:
        def callback(partial_report: str) -> None:
            try:
                mcp = self._build_progress_mcp(
                    exec_request,
                    exec_request.start_time_ms,
                    partial_report=partial_report,
                )
                self._datahub_graph.emit_mcp(mcp, async_flag = False)
            except Exception:
                logger.warning("Failed to emit progress MCP", exc_info=True)

        return callback

    # Builds an MCP to report the start of execution request handling
    def _build_kickoff_mcp(
        self, exec_request: ExecutionRequest
    ) -> MetadataChangeProposalWrapper:
        assert exec_request.exec_id, f"Missing exec_id for request {exec_request}"
        key_aspect = self._build_execution_request_key_aspect(exec_request.exec_id)
        result_aspect = self._build_execution_request_result_aspect(
            status=Type.RUNNING.name, start_time_ms=exec_request.start_time_ms
        )

        return MetadataChangeProposalWrapper(
            entityType=DATAHUB_EXECUTION_REQUEST_ENTITY_NAME,
            changeType="UPSERT",
            auditHeader=None,
            entityKeyAspect=key_aspect,
            aspectName=DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME,
            aspect=result_aspect,
        )

    def _build_progress_mcp(
        self, exec_request: ExecutionRequest, start_time_ms: int, partial_report: str
    ) -> MetadataChangeProposalWrapper:
        assert exec_request.exec_id, f"Missing exec_id for request {exec_request}"
        key_aspect = self._build_execution_request_key_aspect(exec_request.exec_id)
        result_aspect = self._build_execution_request_result_aspect(
            status=Type.RUNNING.name,
            start_time_ms=start_time_ms,
            report=partial_report,
        )

        return MetadataChangeProposalWrapper(
            entityType=DATAHUB_EXECUTION_REQUEST_ENTITY_NAME,
            changeType="UPSERT",
            auditHeader=None,
            entityKeyAspect=key_aspect,
            aspectName=DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME,
            aspect=result_aspect,
        )

    def _build_completion_mcp(
        self, exec_result: ExecutionResult
    ) -> MetadataChangeProposalWrapper:
        # TODO: Support timed out state.

        exec_request = exec_result.context.request
        status = exec_result.type.name
        report = exec_result.get_summary()
        structured_report = exec_result.get_structured_report()

        start_time_ms = exec_request.start_time_ms
        end_time_ms = exec_result.end_time_ms

        key_aspect = self._build_execution_request_key_aspect(
            exec_request.exec_id or "missing execution id"
        )
        result_aspect = self._build_execution_request_result_aspect(
            status=status,
            start_time_ms=start_time_ms,
            duration_ms=end_time_ms - start_time_ms if end_time_ms else None,
            report=report,
            structured_report=StructuredExecutionReportClass(
                type=exec_request.name,
                serializedValue=structured_report,
                contentType="application/json",
            )
            if structured_report
            else None,
        )

        return MetadataChangeProposalWrapper(
            entityType=DATAHUB_EXECUTION_REQUEST_ENTITY_NAME,
            changeType="UPSERT",
            auditHeader=None,
            entityKeyAspect=key_aspect,
            aspect=result_aspect,
        )

    # Builds an MCP to report the completion of execution request handling
    def _build_empty_cancel_mcp(self, exec_id: str) -> MetadataChangeProposalWrapper:
        key_aspect = self._build_execution_request_key_aspect(exec_id)

        # TODO: Determine whether this is the "right" thing to do.
        result_aspect = self._build_execution_request_result_aspect(
            status=Type.CANCELLED.name,
            start_time_ms=0,  # TODO: Make start time optional
            duration_ms=None,
            report="No active execution request found.",
        )

        return MetadataChangeProposalWrapper(
            entityType=DATAHUB_EXECUTION_REQUEST_ENTITY_NAME,
            entityKeyAspect=key_aspect,
            aspect=result_aspect,
        )

    def _build_execution_request_key_aspect(
        self, execution_request_id: str
    ) -> ExecutionRequestKeyClass:
        return ExecutionRequestKeyClass(id=execution_request_id)

    def _build_execution_request_result_aspect(
        self,
        status: str,
        start_time_ms: int,
        duration_ms: Optional[int] = None,
        report: Optional[str] = None,
        structured_report: Optional[StructuredExecutionReportClass] = None,
    ) -> ExecutionRequestResultClass:
        return ExecutionRequestResultClass(
            status=status,
            startTimeMs=start_time_ms,
            durationMs=duration_ms,
            report=report,
            structuredReport=structured_report,
        )

    def _emit_completion_mcp(
        self, exec_id: Optional[str], mcp: MetadataChangeProposalWrapper
    ) -> bool:
        noretry_status = [401, 404, 422]
        try:
            self._datahub_graph.emit_mcp(mcp, async_flag = False)
        except Exception as e:
            if isinstance(e, OperationalError):
                status = e.info.get('status', -1)
                if status in noretry_status:
                    logger.warning(f"Will not retry failed completion MCP for status code {status}")
                    return True
            logger.warning("Failed to emit completion MCP", exc_info=True)
            return False

        if exec_id:
            self.results_to_emit.pop(exec_id, None)
        return True



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datahub.ingestion.run.pipeline import Pipeline

from acryl.executor.common.config import BaseModel, ConfigModel
from acryl.executor.context.execution_context import ExecutionContext
from acryl.executor.context.executor_context import ExecutorContext
from acryl.executor.execution.task import Task


class InMemoryIngestionTaskConfig(ConfigModel):
    config_1: str


class InMemoryIngestionTaskArgs(BaseModel):
    recipe: dict


class InMemoryIngestionTask(Task):
    config: InMemoryIngestionTaskConfig

    @classmethod
    def create(cls, config: dict, ctx: ExecutorContext) -> "Task":
        config = InMemoryIngestionTaskConfig.parse_obj(config)
        return cls(config)

    def __init__(self, config: InMemoryIngestionTaskConfig):
        self.config = config

    def execute(self, args: dict, ctx: ExecutionContext) -> None:
        validated_args = InMemoryIngestionTaskArgs.parse_obj(args)
        pipeline = Pipeline.create(validated_args.recipe)
        pipeline.run()
        pipeline.raise_from_status()

    def close(self) -> None:
        pass



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from dataclasses import dataclass, field
from typing import List, Optional

from datahub.ingestion.api.report import Report


def format_report_line(line_type: str, message: str) -> str:
    utc_time = datetime.datetime.utcnow()
    return f"{utc_time} {line_type}: {message}"


@dataclass
class ExecutionReport(Report):
    """
    Report object leveraged by an Executor coordinating Task Execution.
    """

    exec_id: str

    infos: List[str] = field(default_factory=list)
    errors: List[str] = field(default_factory=list)
    _structured_report: Optional[str] = None
    _logs: str = ""

    def __init__(self, exec_id: str):
        self.exec_id = exec_id
        self.infos = []
        self.errors = []

    def report_info(self, message: str, log: bool = True) -> None:
        formatted_message = format_report_line("INFO", message)
        if log:
            print(f"[exec_id={self.exec_id}] {formatted_message}")
        self.infos.append(formatted_message)

    def report_error(self, message: str, log: bool = True) -> None:
        formatted_message = format_report_line("ERROR", message)
        if log:
            print(f"[exec_id={self.exec_id}] {formatted_message}")
        self.errors.append(formatted_message)

    def set_logs(self, logs: str) -> None:
        self._logs = logs

    def get_logs(self) -> str:
        return self._logs

    def set_structured_report(self, report_content: str) -> None:
        self._structured_report = report_content

    def get_structured_report(self) -> Optional[str]:
        return self._structured_report



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from typing import Callable, Optional, Union

from pydantic import BaseModel

ExecutionProgressCallback = Callable[[str], None]


class ExecutionRequest(BaseModel):
    # Optional executor id to target. Fallbacks back the default (local) executor.
    executor_id: str = "default"

    # Optional run execution id provided by the caller
    exec_id: Union[str, None]

    # The name of the task to be executed
    name: str

    # Arguments to provide the task to be executed
    args: dict

    # Callback to report progress.
    progress_callback: Optional[ExecutionProgressCallback] = None

    # Set by the reporting executor
    start_time: Optional[datetime] = None

    @property
    def start_time_ms(self) -> int:
        if self.start_time:
            return int(self.start_time.timestamp() * 1000)
        else:
            return 0



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import inspect
from typing import Any, Dict, Type, Union

from acryl import __package_name__
from acryl.executor.execution.task import Task


def _is_importable(path: str) -> bool:
    return "." in path or ":" in path


def import_path(path: str) -> Any:
    """
    Import an item from a package, where the path is formatted as 'package.module.submodule.ClassName'
    or 'package.module.submodule:ClassName.classmethod'. The dot-based format assumes that the bit
    after the last dot is the item to be fetched. In cases where the item to be imported is embedded
    within another type, the colon-based syntax can be used to disambiguate.
    """
    assert _is_importable(path), "path must be in the appropriate format"

    if ":" in path:
        module_name, object_name = path.rsplit(":", 1)
    else:
        module_name, object_name = path.rsplit(".", 1)

    item = importlib.import_module(module_name)
    for attr in object_name.split("."):
        item = getattr(item, attr)
    return item


# A task registry is used to store task name -> class mappings for instantiation later.
class TaskRegistry:
    # Map of name to class reference or class type.
    _mapping: Dict[str, Union[str, Type[Task], Exception]]

    def __init__(self) -> None:
        self._mapping = {}

    def _get_registered_type(self) -> Type[Task]:
        return Task

    def _check_cls(self, cls: Type[Task]) -> None:
        if inspect.isabstract(cls):
            raise ValueError(
                f"cannot register an abstract type in the registry; got {cls}"
            )
        super_cls = self._get_registered_type()
        if not issubclass(cls, super_cls):
            raise ValueError(f"must be derived from {super_cls}; got {cls}")

    def _register(
        self, key: str, tp: Union[str, Type[Task], Exception], override: bool = False
    ) -> None:
        if not override and key in self._mapping:
            raise KeyError(f"key already in use - {key}")
        if _is_importable(key):
            raise KeyError(f"key cannot contain special characters; got {key}")
        self._mapping[key] = tp

    def register(self, key: str, cls: Type[Task], override: bool = False) -> None:
        self._check_cls(cls)
        self._register(key, cls, override=override)

    def register_lazy(self, key: str, import_path: str) -> None:
        self._register(key, import_path)

    def register_disabled(
        self, key: str, reason: Exception, override: bool = False
    ) -> None:
        self._register(key, reason, override=override)

    def _ensure_not_lazy(self, key: str) -> Union[Type[Task], Exception]:
        path = self._mapping[key]
        if isinstance(path, str):
            try:
                plugin_class = import_path(path)
                self.register(key, plugin_class, override=True)
                return plugin_class
            except (AssertionError, ModuleNotFoundError, ImportError) as e:
                self.register_disabled(key, e, override=True)
                return e
        else:
            return path

    def is_enabled(self, key: str) -> bool:
        tp = self._mapping[key]
        return not isinstance(tp, Exception)

    def get(self, key: str) -> Type[Task]:
        if _is_importable(key):
            # If the key contains a dot or colon, we treat it as a import path and attempt
            # to load it dynamically.
            clazz = import_path(key)
            self._check_cls(clazz)
            return clazz

        if key not in self._mapping:
            raise KeyError(f"Did not find a registered class for Task with type {key}")

        tp = self._ensure_not_lazy(key)
        if isinstance(tp, ModuleNotFoundError):
            raise EnvironmentError(  # Todo use a better error.
                f"{key} is disabled; try running: pip install '{__package_name__}[{key}]'"  # TODO: Think about hown to handle this case.
            ) from tp
        elif isinstance(tp, Exception):
            raise EnvironmentError(
                f"{key} is disabled due to an error in initialization"
            ) from tp
        else:
            # If it's not an exception, then it's a registered type.
            return tp



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

from acryl.executor.report.execution_report import ExecutionReport
from acryl.executor.request.execution_request import ExecutionRequest


def generate_execution_id() -> str:
    # return new execution id
    return str(uuid.uuid4())


"""
Context object passed down into the Task being executed, when its executed
"""


class ExecutionContext:
    # Execution id
    exec_id: str

    # The original request
    request: ExecutionRequest

    # The report associated with the task
    report: ExecutionReport

    def __init__(self, request: ExecutionRequest) -> None:
        self.request = request
        if request.exec_id is None:
            self.exec_id = generate_execution_id()
        else:
            self.exec_id = request.exec_id
        self.report = ExecutionReport(self.exec_id)

    def get_execution_id(self) -> str:
        return self.exec_id

    def get_report(self) -> ExecutionReport:
        return self.report

    def get_task_name(self) -> str:
        return self.request.name



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from acryl.executor.secret.secret_store import SecretStore

"""
Context object for the executor, passed into a Task on creation
"""


class ExecutorContext:
    # Executor ID
    executor_id: str

    # Secret stores that were configured at boot time
    secret_stores: List[SecretStore]

    def __init__(self, executor_id: str, secret_stores: List[SecretStore] = []) -> None:
        self.executor_id = executor_id
        self.secret_stores = secret_stores

    def get_secret_stores(self) -> List[SecretStore]:
        return self.secret_stores



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import os
import signal
import sys
import tarfile
from asyncio import tasks
from collections import deque
from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional

from acryl.executor.cloud_utils.cloud_copier import CloudCopier
from acryl.executor.cloud_utils.s3_cloud_copier import S3CloudCopier
from acryl.executor.common.config import ConfigModel
from acryl.executor.context.execution_context import ExecutionContext
from acryl.executor.context.executor_context import ExecutorContext
from acryl.executor.execution.sub_process_task_common import (
    SubProcessRecipeTaskArgs, SubProcessTaskUtil)
from acryl.executor.execution.task import Task, TaskError

logger = logging.getLogger(__name__)

ARTIFACTS_DIR_NAME = "artifacts"


class SubProcessIngestionTaskConfig(ConfigModel):
    tmp_dir: str = "/tmp/datahub/ingest"
    log_dir: str = "/tmp/datahub/logs"
    heartbeat_time_seconds: int = 2
    max_log_lines: int = SubProcessTaskUtil.MAX_LOG_LINES
    # The following are optional and only used for uploading logs to S3
    cloud_log_bucket: Optional[str] = os.environ.get("DATAHUB_CLOUD_LOG_BUCKET")
    cloud_log_path: Optional[str] = os.environ.get("DATAHUB_CLOUD_LOG_PATH", "")


class SubProcessIngestionTaskArgs(SubProcessRecipeTaskArgs):
    debug_mode: str = "false"  # Expected values are "true" or "false".


class SubProcessIngestionTask(Task):
    config: SubProcessIngestionTaskConfig
    tmp_dir: str  # Location where tmp files will be written (recipes)
    ctx: ExecutorContext

    @classmethod
    def create(cls, config: dict, ctx: ExecutorContext) -> "Task":
        return cls(SubProcessIngestionTaskConfig.parse_obj(config), ctx)

    def __init__(self, config: SubProcessIngestionTaskConfig, ctx: ExecutorContext):
        self.config = config
        self.tmp_dir = config.tmp_dir
        self.ctx = ctx

    def create_tar_from_dir(self, dir_path: str) -> Optional[Path]:
        logger.info(f"Creating tar archives for {dir_path}")
        base_name = os.path.basename(dir_path)
        has_files = False
        tar_file = Path(dir_path).joinpath(f"{base_name}.tgz")

        # We list dirs here to make sure the tar file itself won't be included in the tar file
        files = os.listdir(dir_path)
        with tarfile.open(tar_file, "w:gz") as tar:
            for item in files:
                item_path = os.path.join(dir_path, item)
                logger.info(f"Added to {base_name}.tgz: {item_path}")
                tar.add(item_path, arcname=item)
                has_files = True

        if not has_files:
            return None

        return tar_file

    def create_tar_archives(
        self, artifacts_path: str, cloud_copier: CloudCopier
    ) -> None:
        # Ensure the artifacts_path is absolute
        artifacts_path = os.path.abspath(artifacts_path)

        # Check if the base path exists and is a directory
        if not os.path.exists(artifacts_path) or not os.path.isdir(artifacts_path):
            raise ValueError(
                f"The provided path '{artifacts_path}' does not exist or is not a directory."
            )
        tars = []
        # Iterate over the items in the base directory
        for item in os.listdir(artifacts_path):
            item_path = os.path.join(artifacts_path, item)
            if os.path.isdir(item_path) and item != "artifacts":
                logger.debug(f"Initiate Creating tar archives for {item_path}")
                tar_file = self.create_tar_from_dir(item_path)
                if tar_file:
                    logger.info(f"Created archive: {tar_file}")
                    tars.append(tar_file)

        # Iterate over the items in the artifacts directory
        for item in os.listdir(os.path.join(artifacts_path, "artifacts")):
            item_path = os.path.join(artifacts_path, "artifacts", item)
            if os.path.isdir(item_path):
                logger.debug(f"Initiate Creating tar archives for {item_path}")
                tar_file = self.create_tar_from_dir(item_path)
                if tar_file:
                    logger.info(f"Created archive: {tar_file}")
                    tars.append(tar_file)

        # Create a single tar archive for the single files in the artifacts directory
        for item in os.listdir(os.path.join(artifacts_path, "artifacts")):
            item_path = os.path.join(artifacts_path, "artifacts", item)
            if os.path.isfile(item_path) and not item.endswith(".tgz"):
                tar_file = Path(item_path).with_suffix(".tgz")
                with tarfile.open(tar_file, "w:gz") as artifacts_tar:
                    # Add files directly under the base directory to the artifacts.tgz
                    artifacts_tar.add(item_path, arcname=item)
                    logger.debug(f"Added to {tar_file}: {item_path}")
                logger.info(f"Created archive: {tar_file}")
                tars.append(tar_file)

        for tar_to_upload in tars:
            try:
                relative_path = str(tar_to_upload).replace(artifacts_path, "")
                cloud_copier.upload(str(tar_to_upload), relative_path)
            except Exception:
                logger.exception(f"Failed to upload {tar_to_upload} to S3")
            finally:
                tar_to_upload.unlink()

    async def execute(self, args: dict, ctx: ExecutionContext) -> None:
        exec_id = ctx.exec_id  # The unique execution id.

        exec_out_dir = f"{self.tmp_dir}/{exec_id}"

        # 0. Validate arguments
        validated_args = SubProcessIngestionTaskArgs.parse_obj(args)

        # 1. Resolve the recipe (combine it with others)
        recipe: dict = SubProcessTaskUtil._resolve_recipe(
            validated_args.recipe, ctx, self.ctx
        )
        plugin: str = SubProcessTaskUtil._get_plugin_from_recipe(recipe)

        # 2. Write recipe file to local FS (requires write permissions to /tmp directory)
        recipe_file_path = SubProcessTaskUtil._write_recipe_to_file(
            exec_out_dir, recipe
        )

        # 3. Spin off subprocess to run the run_ingest.sh script
        debug_mode = validated_args.debug_mode
        command_script: str = "run_ingest.sh"

        stdout_lines: deque[str] = deque(maxlen=self.config.max_log_lines)

        # Create log directory if it doesn't exist
        artifact_output_dir = f"{self.config.log_dir}/{exec_id}"
        mode = 0o755
        (Path(artifact_output_dir) / "executor-logs").mkdir(
            mode, parents=True, exist_ok=True
        )
        Path(artifact_output_dir).joinpath("artifacts").mkdir(
            mode, parents=True, exist_ok=True
        )

        full_log_file = open(
            f"{artifact_output_dir}/executor-logs/ingestion-logs.log", "w"
        )

        report_out_file = f"{artifact_output_dir}/artifacts/ingestion_report.json"

        logger.info(f"Starting ingestion subprocess for exec_id={exec_id} ({plugin})")

        subprocess_env = validated_args.get_combined_env_vars()
        subprocess_env["INGESTION_ARTIFACT_DIR"] = f"{artifact_output_dir}/artifacts"
        # using setdefault allows us to prefer TMPDIR set via ingestion args for debugging
        subprocess_env.setdefault("TMPDIR", exec_out_dir)

        ingest_process = await asyncio.create_subprocess_exec(
            *[
                command_script,
                validated_args.get_venv_name(plugin=plugin),
                validated_args.version,
                plugin,
                self.tmp_dir,
                recipe_file_path,
                report_out_file,
                debug_mode,
            ],
            env=subprocess_env,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
            limit=SubProcessTaskUtil.SUBPROCESS_BUFFER_SIZE,
        )

        # 4. Monitor and report progress.
        most_recent_log_ts: Optional[datetime] = None

        async def _read_output_lines() -> None:
            create_new_line = True
            while True:
                assert ingest_process.stdout

                # We can't use the readline method directly.
                # When the readline method hits a LimitOverrunError, it will
                # discard the line or possibly the entire buffer.
                try:
                    line_bytes = await ingest_process.stdout.readuntil(b"\n")
                except asyncio.exceptions.IncompleteReadError as e:
                    # This happens when we reach the end of the stream.
                    line_bytes = e.partial
                except asyncio.exceptions.LimitOverrunError:
                    line_bytes = await ingest_process.stdout.read(
                        SubProcessTaskUtil.MAX_BYTES_PER_LINE
                    )

                # At this point, if line_bytes is empty, then we're at EOF.
                # If it ends with a newline, then we successfully read a line.
                # If it does not end with a newline, then we hit a LimitOverrunError
                # and it contains a partial line.

                if not line_bytes:
                    logger.info(
                        f"Got EOF from subprocess exec_id={exec_id} - stopping log monitor"
                    )
                    break
                line = line_bytes.decode("utf-8")

                nonlocal most_recent_log_ts
                most_recent_log_ts = datetime.now()

                full_log_file.write(line)

                if create_new_line:
                    stdout_lines.append("")
                    sys.stdout.write(f"[{exec_id} logs] ")
                create_new_line = line.endswith("\n")

                current_line_length = len(stdout_lines[-1])
                if current_line_length < SubProcessTaskUtil.MAX_BYTES_PER_LINE:
                    allowed_length = (
                        SubProcessTaskUtil.MAX_BYTES_PER_LINE - current_line_length
                    )
                    if len(line) > allowed_length:
                        trunc_line = f"{line[:allowed_length]} [...truncated]\n"
                    else:
                        trunc_line = line

                    stdout_lines[-1] += trunc_line
                    sys.stdout.write(trunc_line)
                    sys.stdout.flush()
                else:
                    # If we've already reached the max line length, then we simply ignore the rest of the line.
                    pass

                await asyncio.sleep(0)

        async def _report_progress() -> None:
            while True:
                if ingest_process.returncode is not None:
                    logger.info(
                        f"Detected subprocess return code {ingest_process.returncode}, "
                        f"exec_id={exec_id} - stopping logs reporting"
                    )
                    break

                await asyncio.sleep(self.config.heartbeat_time_seconds)

                # Report progress
                if ctx.request.progress_callback:
                    if most_recent_log_ts is None:
                        report = "No logs yet"
                    else:
                        report = SubProcessTaskUtil._format_log_lines(stdout_lines)
                        current_time = datetime.now()
                        if most_recent_log_ts < current_time - timedelta(minutes=2):
                            message = (
                                f"WARNING: These logs appear to be stale. No new logs have been received since {most_recent_log_ts} ({(current_time - most_recent_log_ts).seconds} seconds ago). "
                                "However, the ingestion process still appears to be running and may complete normally."
                            )
                            report = f"{report}\n\n{message}"

                    # TODO maybe use the normal report field here?
                    logger.debug(f"Reporting in-progress for exec_id={exec_id}")
                    ctx.request.progress_callback(report)

                full_log_file.flush()
                await asyncio.sleep(0)

        async def _process_waiter() -> None:
            await ingest_process.wait()
            logger.info(f"Detected subprocess exited exec_id={exec_id}")

        read_output_task = asyncio.create_task(_read_output_lines())
        report_progress_task = asyncio.create_task(_report_progress())
        process_waiter_task = asyncio.create_task(_process_waiter())

        group = tasks.gather(
            read_output_task, report_progress_task, process_waiter_task
        )
        try:
            await group
        except Exception as e:
            # This could just be a normal cancellation or it could be that
            # one of the monitoring tasks threw an exception.
            # In this case, we should kill the subprocess and cancel the other tasks.
            ingest_process.terminate()

            # If the cause of the exception was a cancellation, then this is a no-op
            # because the gather method already propagates the cancellation.
            group.cancel()

            # ALL_COMPLETED means we wait for all tasks to finish, even if one of them
            # throws an exception. Set timeout to 60s to avoid hanging forever.
            _done, pending = await asyncio.wait(
                (
                    ingest_process.wait(),
                    read_output_task,
                    report_progress_task,
                    process_waiter_task,
                ),
                timeout=60,
                return_when=asyncio.ALL_COMPLETED,
            )
            if pending:
                logger.info(f"Failed to cancel {len(pending)} tasks on cleanup.")
                ingest_process.kill()

            if isinstance(e, asyncio.CancelledError):
                # If it was a cancellation, then we re-raise.
                raise
            else:
                raise RuntimeError(
                    f"Something went wrong in the subprocess executor: {e}"
                ) from e
        finally:
            full_log_file.close()

            if os.path.exists(report_out_file):
                with open(report_out_file, "r") as structured_report_fp:
                    ctx.get_report().set_structured_report(structured_report_fp.read())

            if self.config.cloud_log_bucket:
                upload_time = datetime.now()
                partition = f"year={upload_time.strftime('%Y')}/month={upload_time.strftime('%m')}/day={upload_time.strftime('%d')}"
                try:
                    cloud_copier = S3CloudCopier(
                        self.config.cloud_log_bucket,
                        (self.config.cloud_log_path or "")
                        + "/"
                        + recipe.get("pipeline_name", "unknown_pipeline").replace(
                            "urn:li:dataHubIngestionSource:", ""
                        )
                        + "/"
                        + partition
                        + "/"
                        + ctx.exec_id,
                    )

                    self.create_tar_archives(artifact_output_dir, cloud_copier)
                except Exception:
                    logging.exception("Failed to upload logs to S3")

            ctx.get_report().set_logs(
                SubProcessTaskUtil._format_log_lines(stdout_lines)
            )

            # Cleanup by removing the recipe file
            SubProcessTaskUtil._remove_directory(exec_out_dir)

        return_code = ingest_process.returncode
        if return_code != 0:  # Failed
            if return_code and return_code < 0:
                try:
                    signal_name = signal.Signals(-return_code).name
                except ValueError:
                    signal_name = str(-return_code)
                ctx.get_report().report_error(
                    f"The ingestion process was killed by signal {signal_name} likely because it ran out of memory. "
                    "You can resolve this issue by allocating more memory to the datahub-actions container."
                )
            elif return_code == 137:
                ctx.get_report().report_error(
                    "The ingestion process was terminated with exit code 137, likely because it ran out of memory."
                    "You can resolve this issue by allocating more memory to the datahub-actions container."
                )
            else:
                ctx.get_report().report_info(
                    f"Failed to execute 'datahub ingest', exit code {return_code}"
                )
            raise TaskError("Failed to execute 'datahub ingest'")

        # Report Successful execution
        ctx.get_report().report_info("Successfully executed 'datahub ingest'")

    def close(self) -> None:
        pass



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import json
import logging
import os
import re
import shutil
from typing import Any, List, Sequence, Type

import pydantic
import yaml

from acryl.executor.common.config import PermissiveConfigModel
from acryl.executor.context.execution_context import ExecutionContext
from acryl.executor.context.executor_context import ExecutorContext
from acryl.executor.execution.task import TaskError

logger = logging.getLogger(__name__)


class SubProcessTaskUtil:
    MAX_LOG_LINES = 2000

    # The original value is 64kb (https://github.com/python/cpython/blob/7528e2c06c8baf809b56f406bcc50e8436c9647c/Lib/asyncio/streams.py#L23).
    # Increasing it should improve performance.
    SUBPROCESS_BUFFER_SIZE = 2**20  # 1mb

    # GMS / mysql has a 4mb limit on the size of a data packet.
    # Doing 90% of that so we have some buffer.
    MAX_LOG_SIZE_BYTES = int(0.9 * 2**22)  # 90% of 4mb

    # We want to truncate long lines so that we can show more lines in the logs.
    MAX_BYTES_PER_LINE = 2**12  # 4kb

    @staticmethod
    def _format_log_lines(lines: Sequence[str]) -> str:
        text = "".join(lines)

        # Python slices are super permissive on index bounds, so this works.
        text = text[-SubProcessTaskUtil.MAX_LOG_SIZE_BYTES :]

        if len(lines) >= SubProcessTaskUtil.MAX_LOG_LINES:
            # lines is a deque, so len(lines) won't be larger than MAX_LOG_LINES.
            text = f"[earlier logs truncated...]\n{text}"

        return text

    @staticmethod
    def _resolve_secrets(secret_names: List[str], ctx: ExecutorContext) -> dict:
        # Attempt to resolve secret using by checking each configured secret store.
        secret_stores = ctx.get_secret_stores()
        final_secret_values = dict({})

        for secret_store in secret_stores:
            try:
                # Retrieve secret values from the store.
                secret_values_dict = secret_store.get_secret_values(secret_names)
                # Overlay secret values from each store, if not None.
                for secret_name, secret_value in secret_values_dict.items():
                    if secret_value is not None:
                        final_secret_values[secret_name] = secret_value
            except Exception:
                logger.exception(
                    f"Failed to fetch secret values from secret store with id {secret_store.get_id()}"
                )
        return final_secret_values

    @staticmethod
    def _resolve_recipe(
        recipe: str, execution_ctx: ExecutionContext, executor_ctx: ExecutorContext
    ) -> dict:
        # Now attempt to find and replace all secrets inside the recipe.
        secret_pattern = re.compile(r".*?\${(\w+)}.*?")

        resolved_recipe = recipe
        secret_matches = secret_pattern.findall(resolved_recipe)

        # 1. Extract all secrets needing resolved.
        secrets_to_resolve = []
        if secret_matches:
            for match in secret_matches:
                secrets_to_resolve.append(match)

        # 2. Resolve secret values
        secret_values_dict = SubProcessTaskUtil._resolve_secrets(
            secrets_to_resolve, executor_ctx
        )

        # 3. Substitute secrets into recipe file
        if secret_matches:
            for match in secret_matches:
                # a. Check if secret was successfully resolved.
                secret_value = secret_values_dict.get(match)
                if secret_value is None:
                    # Failed to resolve secret.
                    raise TaskError(
                        f"Failed to resolve secret with name {match}. Aborting recipe execution."
                    )

                # b. Substitute secret value.
                resolved_recipe = resolved_recipe.replace(f"${{{match}}}", secret_value)

        json_recipe = json.loads(resolved_recipe, strict=False)

        # Inject run_id into the recipe
        json_recipe["run_id"] = execution_ctx.exec_id

        # For now expect that the recipe is complete, this may not be the case, however for hybrid deployments. Secret store!
        return json_recipe

    @staticmethod
    def _get_plugin_from_recipe(recipe: dict) -> str:
        # The source type -- ASSUMPTION ALERT: This should always correspond to the plugin name.
        return recipe["source"]["type"]

    @staticmethod
    def _write_recipe_to_file(
        dir_path: str, recipe: dict, file_name: str = "recipe.yml"
    ) -> str:
        # 1. Create directories to the path
        os.makedirs(dir_path, mode=0o777, exist_ok=True)

        # 2. Dump recipe dictionary to a YAML string
        yaml_recipe = yaml.dump(recipe)

        # 3. Write YAML to file
        file_path = dir_path + "/" + file_name
        with open(file_path, "w") as file_handle:
            file_handle.write(yaml_recipe)

        return file_path

    @staticmethod
    def _remove_directory(dir_path: str) -> None:
        shutil.rmtree(dir_path)


def pydantic_parse_json(field: str) -> classmethod:
    def _parse_from_json(cls: Type, v: Any) -> dict:
        if isinstance(v, str):
            return json.loads(v)
        return v

    return pydantic.validator(field, pre=True, allow_reuse=True)(_parse_from_json)


class SubProcessRecipeTaskArgs(PermissiveConfigModel):
    recipe: str
    version: str = "latest"

    extra_pip_requirements: List[str] = []
    extra_pip_plugins: List[str] = []
    extra_env_vars: dict = {}

    _json_extra_pip_requirements = pydantic_parse_json("extra_pip_requirements")
    _json_extra_pip_plugins = pydantic_parse_json("extra_pip_plugins")
    _json_extra_env_vars = pydantic_parse_json("extra_env_vars")

    def get_venv_name(self, plugin: str) -> str:
        # Generate a stable name for the venv.
        # env vars are not included in the hash.
        suffix = hashlib.sha256()
        suffix.update(self.version.encode("utf-8"))
        suffix.update(str(self.extra_pip_requirements).encode("utf-8"))
        suffix.update(str(self.extra_pip_plugins).encode("utf-8"))

        return f"{plugin}-{suffix.digest().hex()[:16]}"

    def get_combined_env_vars(self) -> dict:
        # Combines os.environ, custom PIP_ env vars, and user-provided custom env vars.
        return {
            **os.environ,
            "PIP_EXTRA_PLUGINS": ",".join(self.extra_pip_plugins),
            "PIP_EXTRA_PACKAGES": " ".join(self.extra_pip_requirements),
            **self.extra_env_vars,
        }



import logging

import boto3

from acryl.executor.cloud_utils.cloud_copier import CloudCopier

logger = logging.getLogger(__name__)


class S3CloudCopier(CloudCopier):
    log = logging.getLogger(__name__)

    def __init__(self, bucket: str, base_path: str) -> None:
        self.bucket = bucket
        self.base_path = base_path
        self.session = boto3.session.Session()

    def upload(self, source_local_file: str, target_cloud_file: str) -> None:
        s3 = self.session.resource("s3")

        key = self.base_path.rstrip("/") + "/" + target_cloud_file.lstrip("/")
        logger.info(
            f"Uploading {source_local_file} to bucket: {self.bucket} and base path {self.base_path} and key {key}"
        )
        s3.meta.client.upload_file(
            Filename=source_local_file,
            Bucket=self.bucket,
            Key=key,
        )



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import os
import subprocess
import sys
from collections import deque

from acryl.executor.common.config import ConfigModel
from acryl.executor.context.execution_context import ExecutionContext
from acryl.executor.context.executor_context import ExecutorContext
from acryl.executor.execution.sub_process_task_common import (
    SubProcessRecipeTaskArgs, SubProcessTaskUtil)
from acryl.executor.execution.task import Task, TaskError

logger = logging.getLogger(__name__)


class SubProcessTestConnectionTaskConfig(ConfigModel):
    tmp_dir: str = "/tmp/datahub/ingest"


class SubProcessTestConnectionTaskArgs(SubProcessRecipeTaskArgs):
    pass


class SubProcessTestConnectionTask(Task):
    config: SubProcessTestConnectionTaskConfig
    tmp_dir: str  # Location where tmp files will be written (recipes)
    ctx: ExecutorContext

    @classmethod
    def create(cls, config: dict, ctx: ExecutorContext) -> "Task":
        config_parsed = SubProcessTestConnectionTaskConfig.parse_obj(config)
        return cls(config_parsed, ctx)

    def __init__(
        self, config: SubProcessTestConnectionTaskConfig, ctx: ExecutorContext
    ):
        self.config = config
        self.tmp_dir = config.tmp_dir
        self.ctx = ctx

    async def execute(self, args: dict, ctx: ExecutionContext) -> None:
        exec_id = ctx.exec_id  # The unique execution id.

        exec_out_dir = f"{self.tmp_dir}/{exec_id}"

        # 0. Validate arguments
        validated_args = SubProcessTestConnectionTaskArgs.parse_obj(args)

        # 1. Resolve the recipe (combine it with others)
        recipe: dict = SubProcessTaskUtil._resolve_recipe(
            validated_args.recipe, execution_ctx=ctx, executor_ctx=self.ctx
        )
        plugin: str = SubProcessTaskUtil._get_plugin_from_recipe(recipe)

        # 2. Write recipe file to local FS (requires write permissions to /tmp directory)
        recipe_file_path = SubProcessTaskUtil._write_recipe_to_file(
            exec_out_dir, recipe
        )

        # 3. Spin off subprocess to run the run_ingest.sh script
        command_script: str = "run_test_connection.sh"
        report_out_file: str = f"{exec_out_dir}/connection_report.json"
        stdout_lines: deque = deque(maxlen=SubProcessTaskUtil.MAX_LOG_LINES)

        # TODO: Inject a token into the recipe to run such that it "just works" (currently, a token is required)

        ingest_process = subprocess.Popen(
            [
                command_script,
                validated_args.get_venv_name(plugin=plugin),
                validated_args.version,
                plugin,
                self.tmp_dir,
                recipe_file_path,
                report_out_file,
            ],
            env=validated_args.get_combined_env_vars(),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )

        try:
            while ingest_process.poll() is None:
                assert ingest_process.stdout
                line = ingest_process.stdout.readline()

                sys.stdout.write(line)
                stdout_lines.append(line)
                await asyncio.sleep(0)

            return_code = ingest_process.poll()

        except asyncio.CancelledError:
            # Terminate the running child process
            ingest_process.terminate()
            raise

        finally:
            if os.path.exists(report_out_file):
                with open(report_out_file, "r") as structured_report_fp:
                    ctx.get_report().set_structured_report(structured_report_fp.read())

            ctx.get_report().set_logs(
                SubProcessTaskUtil._format_log_lines(stdout_lines)
            )

            # Cleanup by removing the exec out directory
            SubProcessTaskUtil._remove_directory(exec_out_dir)

        if return_code != 0:
            # Failed
            ctx.get_report().report_info("Failed to execute 'datahub test connection'")
            raise TaskError("Failed to execute 'datahub test connection'")

        # Report Successful execution
        ctx.get_report().report_info("Successfully executed 'datahub test connection'")

    def close(self) -> None:
        pass



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Dict, Set
import setuptools

package_metadata: dict = {}
with open("./src/acryl/__init__.py") as fp:
    exec(fp.read(), package_metadata)


def get_long_description():
    root = os.path.dirname(__file__)
    with open(os.path.join(root, "README.md")) as f:
        description = f.read()

    return description

aws_common = {
    # AWS Python SDK
    "boto3",
    # Deal with a version incompatibility between botocore (used by boto3) and urllib3.
    # See https://github.com/boto/botocore/pull/2563.
    "botocore!=1.23.0",
}

base_requirements = {
    # Compatibility.
    "typing_extensions>=3.7.4; python_version < '3.11'",
    "mypy_extensions>=0.4.3",
    "sqlalchemy-stubs>=0.4",
    # Actual dependencies.
    "pydantic>=1.5.1",
    "acryl-datahub[datahub-rest]>=0.13.3.4rc2",
    *aws_common,
}

plugins: Dict[str, Set[str]] = {
    # Task plugins
    "aws-sqs-remote-executor": aws_common,
    # Secret store plugins
    "aws-sm-secret-store": aws_common,
}

mypy_stubs = {
    "types-dataclasses",
    "types-python-dateutil",
    "types-requests",
    "types-toml",
    "types-PyYAML",
    "types-freezegun",
}

dev_requirements = {
    *base_requirements,
    *mypy_stubs,
    "flake8>=3.8.3",
    "flake8-tidy-imports>=4.3.0",
    "mypy>=0.901,<0.920",
    "pytest>=6.2.2",
    "requests-mock",
    "freezegun",
}

entry_points = {
    # Tasks
    "datahub.executor.task.plugins": [
        "sub_process_ingestion_task = acryl.executor.execution.sub_process_ingestion_task:SubProcessIngestionTask"
    ]
    # todo: add secret store plugins also.
}


setuptools.setup(
    # Package metadata.
    name=package_metadata["__package_name__"],
    version=package_metadata["__version__"],
    url="https://datahubproject.io/",
    project_urls={
        "Documentation": "https://datahubproject.io/docs/",
        "Source": "https://github.com/linkedin/datahub",
        "Changelog": "https://github.com/linkedin/datahub/releases",
    },
    license="Apache License 2.0",
    description="An library used within Acryl Agents to execute tasks",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Unix",
        "Operating System :: POSIX :: Linux",
        "Environment :: Console",
        "Environment :: MacOS X",
        "Topic :: Software Development",
    ],
    # Package info.
    scripts=[
        "scripts/run_ingest.sh",
        "scripts/run_test_connection.sh",
        "scripts/ingestion_common.sh",
        "scripts/install_acryl_datahub.sh",
    ],
    zip_safe=False,
    python_requires=">=3.7",
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="./src"),
    entry_points=entry_points,
    # Dependencies.
    install_requires=list(base_requirements),
    extras_require={"dev": list(dev_requirements)},
)
